"""
NMODL Asset Manager for Arbor and NEURON

  ~ Glues your neurons together!

Manage a local NMODL library that's automatically compiled, loaded, and recompiled
whenever you change your NMODL code or simulator environment.
"""

__version__ = "4.0.2"

from ._fs import get_cache_path as _get_cache_path
from ._glia import Glia, MechId
from .assets import Catalogue, Mod, Package
from .exceptions import *
from .neuron import MechAccessor

_manager = Glia()


def insert(section, asset, variant=None, pkg=None, /, attributes=None, x=None):
    """
    Insert a mechanism or point process into a Section.

    :param section: The section to insert the asset into.
    :type section: Section
    :param asset: The name of the asset. Will be resolved into a fully qualified NEURON name based on preferences, unless a fully qualified name is given.
    :type asset: string
    :param attributes: Attributes of the asset to set on the section/mechanism.
    :type attributes: dict
    :param pkg: Package preference. Overrides global & script preferences.
    :type pkg: string
    :param variant: Variant preference. Overrides global & script preferences.
    :type variant: string
    :param x: Position along the `section` to place the point process at. Does not apply to mechanisms.
    :type x: float
    :raises: LibraryError if the asset isn't found or was incorrectly marked as a point process.
    """
    return _manager.insert(section, asset, variant, pkg, attributes, x)


def resolve(asset, pkg=None, variant=None):
    """
    Resolve an asset selection command to the name known to NEURON.
    """
    return _manager.resolve(asset, pkg=pkg, variant=variant)


def select(asset, pkg=None, variant=None):
    """
    Set script scope preferences for an asset.

    :param asset: Unresolved asset name.
    :type asset: string
    :param pkg: Name of the package to prefer.
    :type pkg: string
    :param variant: Name of the variant to prefer.
    :type variant: string
    """
    return _manager.select(asset, pkg=pkg, variant=variant)


def compile():
    """
    Compile and test all mod files found in all Glia packages.
    """
    return _manager.compile()


def load_library():
    """
    Load the glia library (if it isn't loaded yet).
    """
    return _manager.load_library()


def context(assets=None, pkg=None, variant=None):
    """
    Creates a context that sets glia preferences during a `with` statement.
    """
    return _manager.context(assets=assets, pkg=pkg, variant=variant)


def catalogue(name):
    """
    Load or build an Arbor mechanism catalogue.

    :param name: Name of the Glia installed Arbor catalogue.
    :type name: str
    """
    return _manager.catalogue(name)


def package(name):
    """
    Return package.

    :param name: Name of the installed glia package.
    :type name: str
    """
    return _manager.package(name)


def get_packages():
    """
    Return all installed packages.
    """
    return [*_manager.packages]


def get_cache_path():
    """
    Get the cache path where intermediary files are stored.

    Glia also stores a JSON cache in another directory, which
    you won't find here, but ``glia cache --clear`` can reset.
    """
    return _get_cache_path()



"""
Glia Test Mods

Test collection of NMODL assets.

Glia asset bundle. If the Glia Asset Manager (`nmodl-glia`) is installed, the NMODL assets
in this package will automatically be available in your Glia library for use in the Arbor
and NEURON brain simulation engines.
"""

from pathlib import Path

from glia import Mod, Package

__version__ = "1.0.0"
package = Package(
    "glia_test_mods",
    Path(__file__).resolve().parent,
    mods=[
        Mod("mods/Na__0.mod", "Na", dialect="neuron"),
        Mod("mods/AMPA__0.mod", "AMPA", is_point_process=True, dialect="neuron"),
        Mod(
            "mods/AMPA__1.mod",
            "AMPA",
            variant="1",
            is_point_process=True,
            dialect="neuron",
        ),
        Mod("mods/Na__t.mod", "Na", variant="t", dialect="neuron"),
        Mod("mods/arbor/Cav2_1.mod", "Cav2_1", dialect="arbor"),
    ],
)



import contextlib
import os
import shutil
import subprocess
import typing
from pathlib import Path
from tempfile import TemporaryDirectory, mkdtemp

from . import _mpi
from ._fs import get_cache_path, read_cache, update_cache
from ._hash import get_package_hash, get_package_mods_hash
from .exceptions import *

if typing.TYPE_CHECKING:
    import arbor


SupportedDialect = typing.Union[typing.Literal["arbor"], typing.Literal["neuron"]]


class _ModList(list):
    def __init__(self, package: "Package", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pkg = package
        for item in self:
            if not isinstance(item, Mod):
                raise PackageFileError(f"Package mod '{item}' is not a valid `glia.Mod`.")
            item.set_package(package)

    def __setitem__(self, key, value):
        if not isinstance(value, Mod):
            raise PackageFileError(f"Package mod '{value}' is not a valid `glia.Mod`.")
        super().__setitem__(key, value)
        value.set_package(self.pkg)

    def append(self, item):
        item.set_package(self.pkg)
        super().append(item)

    def extend(self, itr):
        super().extend(i.set_package(self.pkg) or i for i in itr)


class Package:
    def __init__(self, name: str, root: Path, *, mods: list["Mod"] = None, builtin=False):
        self._name = name
        self._root = Path(root)
        self.mods: list["Mod"] = _ModList(self, [] if mods is None else mods)
        # Exceptional flag for the NEURON builtins.
        # They need a definition to be `insert`ed,
        # but have no mod files to be compiled.
        self.builtin = builtin

    @property
    def name(self):
        return self._name

    @property
    def catalogue(self):
        return Catalogue(self)

    @property
    def hash(self):
        return get_package_hash(self)

    @property
    def mod_hash(self):
        return get_package_mods_hash(self)

    @property
    def root(self):
        return self._root

    def load_catalogue(self) -> "arbor.catalogue":
        return self.catalogue.load()

    def build_catalogue(self, *args, **kwargs) -> "arbor.catalogue":
        return self.catalogue.build(*args, **kwargs)

    def get_mods(self, dialect=None) -> list["Mod"]:
        # Find common mods
        mods = {mod.mod_name: mod for mod in self.mods if mod.dialect is None}
        if dialect:
            # Overwrite with dialect-specific mods
            mods.update(
                (mod.mod_name, mod) for mod in self.mods if mod.dialect == dialect
            )
        return [*mods.values()]


class Mod:
    def __init__(
        self,
        relpath: str,
        asset_name,
        *,
        variant="0",
        is_point_process=False,
        is_artificial_cell=False,
        dialect: SupportedDialect = None,
        builtin=False,
    ):
        self._pkg: typing.Optional[Package] = None
        self.relpath = relpath
        self.asset_name = asset_name
        self.variant = variant
        self.is_point_process = is_point_process
        self.is_artificial_cell = is_artificial_cell
        self.dialect = dialect
        self.builtin = builtin

    def set_package(self, package: Package):
        self._pkg = package

    @property
    def pkg(self):
        return self._pkg

    @property
    def pkg_name(self):
        return self._pkg.name

    @property
    def mech_id(self):
        return ModName(self.pkg_name, self.asset_name, self.variant).mech_id

    @property
    def mod_name(self):
        if self.builtin:
            return self.asset_name
        return ModName(self.pkg_name, self.asset_name, self.variant).full_mod_name

    @property
    def arbor_name(self):
        return ModName(self.pkg_name, self.asset_name, self.variant).arbor_mod_name

    @property
    def path(self) -> Path:
        return self.pkg.root / self.relpath


class ModName:
    def __init__(self, pkg_name: str, asset: str, variant: str):
        self.pkg = pkg_name
        self.asset = asset
        self.variant = variant

    @property
    def full_mod_name(self):
        if self.pkg is None:
            raise ValueError("Missing pkg info for mod name.")
        else:
            return f"glia__{self.pkg}__{self.asset}__{self.variant}"

    @property
    def short_mod_name(self):
        return f"{self.asset}__{self.variant}"

    @property
    def arbor_mod_name(self):
        name = self.asset
        if self.variant and self.variant != "0":
            name += "_" + self.variant
        return name

    @classmethod
    def parse(cls, name: str):
        name_parts = name.split("__")
        if len(name_parts) == 2:
            return cls(None, *name_parts)
        elif len(name_parts) == 4:
            return cls(*name_parts[1:])
        else:
            raise ValueError(f"Unparsable mod name '{name}'.")

    @classmethod
    def parse_path(cls, path: str):
        return cls.parse(Path(path).stem)

    @property
    def mech_id(self):
        if not self.variant:
            return self.asset
        elif not self.pkg:
            return (self.asset, self.variant)
        else:
            return (self.asset, self.variant, self.pkg)


class Catalogue:
    def __init__(self, package: Package):
        self._pkg = package
        self._cache = get_cache_path(self.name, prefix="arb_")

    @property
    def name(self):
        return self._pkg.name

    def load(self) -> "arbor.catalogue":
        import arbor

        if not self.is_fresh():
            self.build()
        return arbor.load_catalogue(self._get_library_path())

    def _get_library_path(self):
        return os.path.join(self._cache, f"{self.name}-catalogue.so")

    def is_fresh(self):
        if not os.path.exists(self._get_library_path()):
            return False
        try:
            cache_data = read_cache()
            cached = cache_data.get("cat_hashes", {}).get(self.name, None)
            return cached == self._hash()
        except FileNotFoundError as _:
            return False

    def _hash(self):
        import arbor

        arbor_hash = str(arbor.config())
        return self._pkg.mod_hash + arbor_hash

    def build(self, verbose=None, debug=False, gpu=None):
        # Turn verbosity on if debug is on, unless it's explicitly toggled off.
        verbose = False if verbose is False else verbose or debug

        def run_build():
            self._build_local(verbose, debug, gpu)

        build_err = None
        try:
            if _mpi.main_node:
                run_build()
        except Exception as err:
            _mpi.bcast(err)
            raise err from None
        else:
            build_err = _mpi.bcast(build_err)
            if build_err:
                raise BuildCatalogueError(
                    "Catalogue build error, look for main node error."
                ) from None

    def _build_local(self, verbose, debug, gpu):
        tmp_dir = TemporaryDirectory
        if debug:
            # Temp dir context manager that doesn't clean up the build folder so that
            # the generated cpp code can be debugged
            class TmpDir:
                def __enter__(*args, **kwargs):
                    return mkdtemp()

                def __exit__(*args, **kwargs):
                    pass

            tmp_dir = TmpDir

        with self.assemble_arbor_mod_dir() as mod_path:
            with tmp_dir() as tmp:
                pwd = os.getcwd()
                os.chdir(tmp)
                try:
                    cmd = f"arbor-build-catalogue {self.name} {mod_path}"
                    subprocess.run(
                        cmd
                        + (" --quiet" if not verbose else "")
                        + (" --verbose" if verbose else "")
                        + (" --debug" if debug else "")
                        + (f" --gpu={gpu}" if gpu else ""),
                        shell=True,
                        check=True,
                        capture_output=not verbose,
                    )
                except subprocess.CalledProcessError as e:
                    msg_p = [f"ABC errored out with exitcode {e.returncode}"]
                    if verbose:
                        msg_p += ["Check log above for error."]
                    else:
                        msg_p += [
                            f"Command: {cmd}\n---- ABC output ----",
                            e.stdout.decode(),
                            "---- ABC error  ----",
                            e.stderr.decode(),
                        ]
                    msg = "\n\n".join(msg_p)
                    raise BuildCatalogueError(msg) from None
                else:
                    os.makedirs(self._cache, exist_ok=True)
                    shutil.copy2(f"{self.name}-catalogue.so", self._cache)
                finally:
                    os.chdir(pwd)
        if debug:
            print(f"Debug copy of catalogue in '{tmp}'")
        # Cache directory hash of current mod files so we only rebuild on source code
        # changes.
        cache_data = read_cache()
        cat_hashes = cache_data.setdefault("cat_hashes", dict())
        cat_hashes[self.name] = self._hash()
        update_cache(cache_data)

    @contextlib.contextmanager
    def assemble_arbor_mod_dir(self):
        from .packaging import NmodlWriter

        with TemporaryDirectory() as tmpdir:
            for mod in self._pkg.get_mods(dialect="arbor"):
                writer = NmodlWriter(mod)
                writer.parse_source(mod.path)
                # `arbor-build-catalogue` needs filename to match suffix statement.
                # See https://github.com/arbor-sim/arbor/issues/2250
                writer.update_suffix_ast(mod.arbor_name)
                writer.write(
                    (Path(tmpdir) / mod.arbor_name).with_suffix(".mod"), dialect="arbor"
                )
            yield tmpdir



import unittest
from pathlib import Path

from glia._fs import get_cache_path, log


class TestLog(unittest.TestCase):
    def test_log_hash(self):
        # We use the fact that Python salts hashes per session (see
        # https://docs.python.org/3/using/cmdline.html#envvar-PYTHONHASHSEED) to get a
        # a unique logfile hash per process, test this.
        self.assertEqual(id("5"), id("5"))

    def test_log(self):
        log_path = Path(get_cache_path(f"{id('5')}.txt"))
        log_path.unlink(missing_ok=True)
        log("hello world")
        self.assertTrue(log_path.exists(), "Logs not created")
        self.assertIn("hello world", log_path.read_text(), "Log not logged")

    def test_exc(self):
        log_path = Path(get_cache_path(f"{id('5')}.txt"))
        log_path.unlink(missing_ok=True)
        try:
            raise RuntimeError()
        except RuntimeError as e:
            log("hello world", exc=e)
        self.assertTrue(log_path.exists(), "Logs not created")
        # Check log level elevation to ERROR
        self.assertIn("ERROR] hello world", log_path.read_text(), "Exc not logged")



"""
    Resolves package, mechanism and variant constraints into asset names that can be
    requested from the Glia library.
"""

from contextlib import contextmanager

from ._fs import read_preferences, write_preferences
from .exceptions import (
    AssetLookupError,
    NoMatchesError,
    TooManyMatchesError,
    UnknownAssetError,
)


class IndexEntry:
    def __init__(self, name):
        self.name = name
        self.items = []

    def __call__(self):
        return self.items

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def append(self, asset):
        self.items.append(asset)


class _NoPreferenceError(Exception):
    pass


class Resolver:
    def __init__(self, manager):
        self._manager = manager
        self.global_preferences = read_preferences()
        self.local_preferences = {}
        self.__preference_stack = {}
        self.__next_stack_id = 0
        self.construct_index()

    def construct_index(self):
        packages = self._manager.packages
        self.index = {}
        self._reverse_lookup = {}
        for pkg in packages:
            for mod in pkg.mods:
                if mod.dialect is not None and mod.dialect != "neuron":
                    continue
                name = mod.asset_name
                if not name in self.index:
                    self.index[name] = IndexEntry(name)
                self.index[name].append(mod)
                self._reverse_lookup[mod.mod_name] = mod

    def resolve(self, asset_name, pkg=None, variant=None):
        if isinstance(asset_name, tuple):
            try:
                variant = asset_name[1]
                pkg = asset_name[2]
            except IndexError:
                pass
            asset_name = asset_name[0]
        if not asset_name in self.index:
            raise UnknownAssetError(
                "Selection could not be resolved: Asset '{}' not found.".format(
                    asset_name
                )
            )

        # Try resolving with preference
        resolved = self.resolve_preference(asset_name, pkg=pkg, variant=variant)
        if resolved is not None:
            return resolved.mod_name

        # If there was no preference, or if the preference couldn't resolve, try resolving
        # without preference.
        resolved = self._get_resolved(asset_name, pkg, variant)

        if not resolved:
            raise NoMatchesError(
                f"Selection {pkg}.{asset_name}.{variant} could not be resolved.",
                pkg,
                variant,
            )
        elif len(resolved) > 1:
            return self._resolve_multi(resolved, asset_name, pkg, variant).mod_name
        else:
            return resolved[0].mod_name

    def resolve_preference(self, asset_name, pkg=None, variant=None):
        # Combine global, script, context & local preferences to obtain package and variant.
        try:
            pkg, variant = self._get_final_pv(asset_name, pkg, variant)
        except _NoPreferenceError:
            # No preferences found.
            return None

        # Resolve the preferred package and variant
        resolved = self._get_resolved(asset_name, pkg, variant)
        if len(resolved) == 1:
            # Exactly 1 candidate found. Return it.
            return resolved[0]
        elif len(resolved) > 1:
            # Multiple candidates found, check if one obvious candidate stands out,
            # if not throw TooManyMatchesError.
            return self._resolve_multi(resolved, asset_name, pkg, variant)
        else:
            # No candidates found.
            return None

    def lookup(self, mod_name):
        if mod_name not in self._reverse_lookup:
            raise AssetLookupError("No mod with name '{}' found".format(mod_name))
        else:
            return self._reverse_lookup[mod_name]

    def _get_final_pv(self, asset_name, pkg, variant):
        """
        Get final package and variant preference.

        This function fetches all preferences and checks whether there's specific
        package or variant preferences. If not it checks the general package or
        variant preferences not specific to the asset.
        """
        # Fetch global, script & context preferences
        preferences = self._preferences()
        if not asset_name in preferences:
            # No specific asset preferences found, but maybe general package or variant
            # preferences apply?
            if self._has_general_preferences(preferences):
                # Get the general package and variant preferences
                p_pkg, p_variant = self._general_preferences(preferences, pkg, variant)
                if pkg is None:
                    pkg = p_pkg
                if variant is None:
                    variant = p_variant
            else:
                # No preferences of any sort found.
                raise _NoPreferenceError()
        else:
            preference = preferences[asset_name]
            if pkg is None and "package" in preference:
                pkg = preference["package"]
            if variant is None and "variant" in preference:
                variant = preference["variant"]
        return pkg, variant

    def _get_resolved(self, asset_name, pkg, variant):
        mods = iter(self.index[asset_name])
        if pkg:
            mods = filter(lambda m: m.pkg_name == pkg, mods)
        if variant:
            mods = filter(lambda m: m.variant == variant, mods)
        return list(mods)

    def _has_general_preferences(self, preferences):
        return ("__pkg" in preferences and preferences["__pkg"]) or (
            "__variant" in preferences and preferences["__variant"]
        )

    def _general_preferences(self, preferences, pkg, variant):
        return (
            pkg or (("__pkg" in preferences and preferences["__pkg"]) or None),
            variant
            or (("__variant" in preferences and preferences["__variant"]) or None),
        )

    def _resolve_multi(self, resolved, asset_name, pkg, variant):
        # If all candidates are from the same package, and 1 is the default variant
        # then return that default variant
        if all(m.pkg_name == resolved[0].pkg_name for m in resolved) and any(
            m.variant == "0" for m in resolved
        ):
            return next(m for m in resolved if m.variant == "0")
        selection = asset_name
        if pkg:
            selection = pkg + "." + asset_name
        if variant:
            selection += " ({})".format(variant)
        raise TooManyMatchesError(
            f"Selection could not be resolved, too many matches for {selection}:\n  * "
            + "\n  * ".join(r.mod_name for r in resolved)
            + "\n Try specifying a package or variant",
            list(r.mod_name for r in resolved),
            asset_name,
            pkg,
            variant,
        )

    def has_preference(self, asset_name):
        return asset_name in self._preferences()

    def set_preference(self, asset_name, glbl=False, pkg=None, variant=None):
        preference = {}
        if pkg is not None:
            preference["package"] = pkg
        if variant is not None:
            preference["variant"] = variant
        if glbl:
            self.global_preferences[asset_name] = preference
            write_preferences(self.global_preferences)
        else:
            self.local_preferences[asset_name] = preference

    @contextmanager
    def preference_context(self, assets=None, pkg=None, variant=None):
        if assets is None:
            assets = {}
        if pkg is not None:
            assets["__pkg"] = pkg
        if variant is not None:
            assets["__variant"] = variant
        id = self._stack_preference_context(assets)
        try:
            yield
        finally:
            self._pop_preference_context(id)

    def _stack_preference_context(self, assets):
        id = self.__next_stack_id
        self.__next_stack_id += 1
        self.__preference_stack[id] = assets
        return id

    def _pop_preference_context(self, id):
        del self.__preference_stack[id]

    def _preferences(self):
        pref = self.global_preferences.copy()
        pref.update(self.local_preferences)
        pref = self._apply_preference_stack(pref)
        return pref

    def _apply_preference_stack(self, base_pref):
        def combine(asset, old, new):
            if asset == "__pkg" or asset == "__variant":
                return new or old
            if "unset" in new and new["unset"]:
                old["package"] = None
                old["variant"] = None
            if "package" in new and new["package"] is not None:
                old["package"] = new["package"]
            if "variant" in new and new["variant"] is not None:
                old["variant"] = new["variant"]
            return old

        compiled_preferences = {}
        for stacked_preference in self.__preference_stack.values():
            for asset, preference in stacked_preference.items():
                if asset not in compiled_preferences:
                    if asset == "__pkg" or asset == "__variant":
                        compiled_preferences[asset] = preference
                    else:
                        compiled_preferences[asset] = {"package": None, "variant": None}
                compiled_preferences[asset] = combine(
                    asset, compiled_preferences[asset], preference
                )
        base_pref.update(compiled_preferences)
        return base_pref



import unittest

import glia._mpi
from glia import PackageError

skipParallel = unittest.skipIf(glia._mpi.parallel_run, "Skip during parallel tests.")


def skipUnlessInstalled(package_name: str):
    try:
        glia.package(package_name)
        installed = True
    except PackageError:
        installed = False
    return unittest.skipUnless(
        installed, f"Skipped because '{package_name}' is not installed."
    )


def skipIfInstalled(package_name: str = None):
    if package_name is None:
        return unittest.skipIf(
            glia.get_packages(), "Skipped because there are installed packages."
        )
    try:
        glia.package(package_name)
        installed = True
    except PackageError:
        installed = False
    return unittest.skipIf(installed, f"Skipped because '{package_name}' is installed.")


skipUnlessTestMods = skipUnlessInstalled("glia_test_mods")



import datetime
import json
import os
import sys
import typing
import warnings
from pathlib import Path
from traceback import format_exception

import appdirs

from glia._hash import hash_path

_install_dirs = appdirs.AppDirs(appname="Glia", appauthor="DBBS")

LogLevel = typing.Union[
    typing.Literal["log"], typing.Literal["warn"], typing.Literal["error"]
]


def log(message: str, *, level: LogLevel = None, category=None, exc: Exception = None):
    log_path = Path(get_cache_path(f"{id('5')}.txt"))
    if exc is not None and level is None:
        level = "error"
    if level:
        level = level.upper()
    header = " ".join(str(c) for c in (datetime.datetime.now(), level, category) if c)
    log_path.parent.mkdir(parents=True, exist_ok=True)
    with open(log_path, "a") as f:
        f.write(f"[{header}] {message}")
        if exc:
            f.write(":\n")
            f.writelines(
                "  " + "\n  ".join(line.split("\n"))
                for line in format_exception(type(exc), exc, exc.__traceback__)
            )
            f.write("Exception arguments:\n")
            f.writelines(f"  {a}\n" for a in exc.args)
            warnings.warn(message + f". See the full log at '{log_path}'.")
        else:
            f.write("\n")


def get_glia_path():
    from . import __path__

    return __path__[0]


def get_cache_hash(prefix=""):
    return prefix + hash_path(get_glia_path())[:8] + hash_path(sys.prefix)[:8]


def get_cache_path(*subfolders, prefix=""):
    return os.path.join(
        _install_dirs.user_cache_dir,
        get_cache_hash(prefix),
        *subfolders,
    )


def get_data_path(*subfolders):
    return os.path.join(_install_dirs.user_data_dir, *subfolders)


def get_neuron_mod_path(*paths):
    return get_cache_path(*paths)


def get_local_pkg_path():
    from . import __version__

    return get_data_path(__version__.split(".")[0], "local")


def _read_shared_storage(*path):
    _path = get_data_path(*path)
    try:
        with open(_path, "r") as f:
            return json.load(f)
    except (IOError, json.JSONDecodeError):
        return {}


def _write_shared_storage(data, *path):
    _path = get_data_path(*path)
    with open(_path, "w") as f:
        f.write(json.dumps(data))


def read_storage(*path):
    data = _read_shared_storage(*path)
    glia_path = get_glia_path()
    if glia_path not in data:
        return {}
    return data[glia_path]


def write_storage(data, *path):
    _path = get_data_path(*path)
    glia_path = get_glia_path()
    shared_data = _read_shared_storage(*path)
    shared_data[glia_path] = data
    _write_shared_storage(shared_data, *path)


def read_cache():
    cache = read_storage("cache.json")
    if "mod_hashes" not in cache:
        cache["mod_hashes"] = {}
    return cache


def write_cache(cache_data):
    write_storage(cache_data, "cache.json")


def update_cache(cache_data):
    cache = read_cache()
    cache.update(cache_data)
    write_cache(cache)


def clear_cache():
    empty_cache = {"mod_hashes": {}, "cat_hashes": {}}
    write_cache(empty_cache)


def read_preferences():
    return read_storage("preferences.json")


def write_preferences(preferences):
    write_storage(preferences, "preferences.json")


def create_preferences():
    write_storage({}, "preferences.json")



import importlib.util
import os
import shutil
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

from click import ClickException
from click.testing import CliRunner

import glia._cli
import glia._fs
import glia._mpi
from glia import Mod, Package, get_cache_path

from . import _shared


def run_cli_command(command, xfail=False, **kwargs):
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(glia._cli.glia, command, **kwargs)
        if (
            not xfail
            and result.exception
            and not isinstance(result.exception, (SystemExit, ClickException))
        ):
            raise result.exception
        return result


def import_tmp_package(path):
    path = Path(path) / "__init__.py"
    spec = importlib.util.spec_from_file_location("__test_package__", path)
    m = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(m)
    return m.package


class TestCLI(unittest.TestCase):
    """
    Test CLI commands
    """

    def test_glia(self):
        result = run_cli_command([])
        self.assertNotEqual("", result.output, "Should show help message")
        self.assertEqual(0, result.exit_code)

    @_shared.skipParallel
    @_shared.skipUnlessTestMods
    def test_compile(self):
        result = run_cli_command(["compile"])
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r"(\d+) out of (\1) passed")

    @_shared.skipParallel
    @_shared.skipIfInstalled()
    def test_compile_nopkg(self):
        result = run_cli_command(["compile"])
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r"0 out of 0 passed")

    def test_list(self):
        result = run_cli_command(["list"])
        self.assertEqual(0, result.exit_code)

    @_shared.skipUnlessTestMods
    def test_show(self):
        result = run_cli_command(["show", "AMPA"])
        self.assertEqual(0, result.exit_code)

    def test_show_unknown(self):
        result = run_cli_command(["show", "doesntexist"], xfail=True)
        self.assertIn('Unknown ASSET "doesntexist"', result.output)
        self.assertEqual(2, result.exit_code)

    @_shared.skipUnlessTestMods
    def test_show_pkg(self):
        # todo: test this test
        result = run_cli_command(["show-pkg", "glia_test_mods"])
        self.assertEqual(0, result.exit_code)

    def test_show_unknown_pkg(self):
        result = run_cli_command(["show-pkg", "doesntexist"], xfail=True)
        self.assertIn('Unknown PACKAGE "doesntexist"', result.output)
        self.assertEqual(2, result.exit_code)

    @_shared.skipUnlessTestMods
    def test_test(self):
        result = run_cli_command(["test", "Na"])

    @_shared.skipUnlessTestMods
    def test_test_unknown(self):
        result = run_cli_command(["test", "unknown"], xfail=True)
        self.assertIn("[?] unknown", result.output)
        self.assertEqual(1, result.exit_code)

    def test_cache(self):
        result = run_cli_command(["cache"])
        self.assertIn(
            glia._fs._install_dirs.user_cache_dir, result.output, "wrong cache dir"
        )

    @_shared.skipParallel
    def test_cache_clear(self):
        pre_test = glia._fs.read_cache()
        with TemporaryDirectory() as tmp:
            shutil.copytree(get_cache_path(), tmp, dirs_exist_ok=True)
            try:
                glia._fs.write_cache({"test_key": "test_value"})
                result = run_cli_command(["cache", "--clear"])
                self.assertFalse(Path(get_cache_path()).exists(), "cache dir not removed")
                self.assertNotIn("test_key", glia._fs.read_cache(), "cache not cleared")
            finally:
                glia._fs.write_cache(pre_test)
                shutil.copytree(tmp, get_cache_path(), dirs_exist_ok=True)


class TestPackagingCLI(unittest.TestCase):
    def test_new_package(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(glia._cli.glia, ["pkg", "new"], input="\n\n\n\n\n\n\n")
            self.assertEqual(
                ["glia_package"],
                os.listdir(),
                "Should have created the package folder",
            )
            self.assertIn(
                "glia_package",
                os.listdir("glia_package"),
                "Should have created the module folder",
            )
            os.chdir("glia_package")
            package = import_tmp_package("glia_package")
            self.assertEqual(Package, type(package))
            self.assertEqual(0, result.exit_code)

    def test_add_mod(self):
        mod_path = Path(__file__).parent / "data" / "mods" / "Na__0.mod"
        mod_source = mod_path.read_text()
        assert "SUFFIX Na\n" in mod_path.read_text(), "Source of Na was mutated"
        runner = CliRunner()
        with runner.isolated_filesystem():
            runner.invoke(glia._cli.glia, ["pkg", "new"], input="\n\n\n\n\n\n\n")
            os.chdir("./glia_package")
            result = runner.invoke(
                glia._cli.glia,
                ["pkg", "add", str(mod_path)],
            )

            self.assertEqual(0, result.exit_code)
            self.assertEqual(
                [".gitkeep", "Na__0.mod"], sorted(os.listdir("glia_package/mods"))
            )
            package = import_tmp_package("glia_package")
            self.assertEqual(Package, type(package), "Broken package declaration")
            self.assertEqual(1, len(package.mods), "Should have added mod")
            mod = package.mods[0]
            mod: Mod
            self.assertEqual(Mod, type(mod), "Broken mod declaration")
            self.assertEqual("Na", mod.asset_name, "Wrong name")
            self.assertEqual("0", mod.variant, "Wrong variant")
            self.assertEqual("glia_package", mod.pkg_name, "Wrong package name")
            self.assertEqual(
                "glia_package/mods/Na__0.mod",
                str(mod.path.relative_to(Path().resolve())),
                "Wrong mod path",
            )
            self.assertTrue(
                mod.path.exists(),
                "Mod path not created",
            )
            mod_text = mod.path.read_text()
            self.assertIn(f"SUFFIX {mod.mod_name}", mod_text, "Mod suffix not written")
            self.assertEqual(
                mod_source, mod_path.read_text(), "Adding mod file changed mod source"
            )

    def test_add_synapse(self):
        mod_path = Path(__file__).parent / "data" / "mods" / "AMPA__0.mod"
        mod_source = mod_path.read_text()
        assert (
            "POINT_PROCESS AMPA\n" in mod_path.read_text()
        ), "Source of AMPA was mutated"
        runner = CliRunner()
        with runner.isolated_filesystem():
            runner.invoke(glia._cli.glia, ["pkg", "new"], input="\n\n\n\n\n\n\n")
            os.chdir("./glia_package")
            result = runner.invoke(
                glia._cli.glia,
                [
                    "pkg",
                    "add",
                    str(mod_path),
                ],
            )

            self.assertEqual(0, result.exit_code)
            package = import_tmp_package("glia_package")
            mod = package.mods[0]
            mod: Mod
            mod_text = mod.path.read_text()
            self.assertIn(
                f"POINT_PROCESS {mod.mod_name}", mod_text, "Mod suffix not written"
            )
            self.assertEqual(
                mod_source, mod_path.read_text(), "Adding mod file changed mod source"
            )



import sys
import unittest


@unittest.skipIf(sys.platform == "win32", "Skip Arbor on windows")
class TestCatalogues(unittest.TestCase):
    """
    Check if catalogues can be built.
    """

    def test_build(self):
        # todo: add catalogue to glia_test_mods, then try to build it here.
        pass



import unittest

from glia.assets import Mod, Package
from glia.resolution import Resolver


class ManagerMock:
    def __init__(self, packages):
        self.packages = packages


class TestResolver(unittest.TestCase):
    """
    Check if packages can be discovered.
    """

    def test_resolve(self):
        pkg = Package(
            "test", __file__, mods=[m := Mod("./doesntexist", "hello", variant="test_v")]
        )
        mname = m.mod_name
        pkg.mods = [m]
        resolver = Resolver(ManagerMock([pkg]))
        self.assertEqual(mname, resolver.resolve("hello"))
        self.assertEqual(mname, resolver.resolve("hello", pkg="test", variant="test_v"))
        # Test the tuple forms
        self.assertEqual(mname, resolver.resolve(("hello",)))
        self.assertEqual(mname, resolver.resolve(("hello", "test_v", "test")))



import abc
import ast
import inspect
import itertools
import re
import typing
from collections import defaultdict
from copy import copy
from pathlib import Path

import black

from ..assets import Mod, Package, SupportedDialect
from ..exceptions import ModSourceError, PackageApiError, PackageFileError

if typing.TYPE_CHECKING:
    from nmodl import NmodlDriver


class ModuleReferences:
    def __init__(self, module: ast.Module):
        self.module_names = []
        self.pkg_names = []
        self.mod_node = None
        self.mod_names = []
        for node in module.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    if alias.name == "glia":
                        self.module_names.append(alias.asname or alias.name)
            elif isinstance(node, ast.ImportFrom) and node.module == "glia":
                for alias in node.names:
                    if alias.name == "Mod":
                        self.mod_names.append(alias.asname or alias.name)
                    if alias.name == "Package":
                        self.pkg_names.append(alias.asname or alias.name)
                        # Store this node to insert `Mod` import if it's missing
                        self.mod_node = node

        if not (self.module_names or self.pkg_names or self.mod_names):
            raise PackageFileError("Package file is missing top-level `glia` imports.")

    def _is_call(self, node: ast.expr, attr, aliases):
        if isinstance(node, ast.Call):
            func = node.func
            if isinstance(func, ast.Name) and func.id in aliases:
                # Class(...)
                #   or
                # ClassAlias(...)
                return True
            elif (
                isinstance(func, ast.Attribute)
                and func.attr == attr
                and isinstance(func.value, ast.Name)
                and func.value.id in self.module_names
            ):
                # glia.Class(...)
                #   or
                # gliaAlias.Class(...)
                return True
        return False

    def is_package_target(self, node: ast.expr):
        return self._is_call(node, "Package", self.pkg_names)

    def is_mod_target(self, node: ast.expr):
        return self._is_call(node, "Mod", self.mod_names)

    def make_mod_call_node(self):
        if self.mod_names:
            func = ast.Name(self.mod_names[0])
        else:
            if not self.module_names:
                # The package file does not contain a `Mod` import, and also no `import
                # glia` so it must contain a `from glia import Package`, so we can
                # append `, Mod`.
                self.mod_node.names.append(ast.alias("Mod"))
                self.mod_names.append("Mod")
                func = ast.Name("Mod")
            else:
                func = ast.Attribute(ast.Name(self.module_names[0]), ast.Name("Mod"))
        return ast.Call(func, [], [])

    def has_mod_ref(self):
        return self.mod_names or self.module_names


class PackageTransformer(ast.NodeTransformer):
    def __init__(self, path: Path, module: ast.Module, pkg_id: str):
        super().__init__()
        self._path = path
        self._module = module
        self._refs = ModuleReferences(self._module)
        self.pkg_node = self.get_package_declaration(pkg_id)

    @property
    def path(self):
        return self._path

    def get_package_declaration(self, pkg_id: str) -> ast.Call:
        pkg = None
        for node in self._module.body:
            if isinstance(node, ast.Assign):
                pkgs = [
                    n for n in node.targets if isinstance(n, ast.Name) and n.id == pkg_id
                ]
                if pkgs:
                    if pkg is not None:
                        raise PackageFileError(
                            f"Multiple top-level '{pkg_id}' assignments."
                        )
                    if len(node.targets) != 1:
                        raise PackageFileError(
                            "Tuple unpacking is not allowed "
                            "in package assigment statements."
                        )
                    if self._refs.is_package_target(node.value):
                        pkg = node.value
                    else:
                        raise PackageFileError(
                            f"Top-level assignment to '{pkg_id}' "
                            "is not a `glia.Package` declaration."
                        )
        if not pkg:
            raise PackageFileError(
                f"No top-level package declaration found for '{pkg_id}'"
            )
        return pkg

    def get_package_shim(self):
        return Package(self.get_package_name(), self._path.parent)

    def get_package_name(self):
        kw_nodes = {kw.arg: kw.value for kw in self.pkg_node.keywords}
        name_node = kw_nodes.get("name")
        if name_node is None:
            try:
                name_node = self.pkg_node.args[0]
            except:
                pass
        if name_node is None:
            raise PackageFileError(
                f"Can't find name argument in package node {ast.unparse(self.pkg_node)}"
            )
        if not isinstance(name_node, ast.Constant):
            raise PackageFileError(f"Package name argument must be a constant.")
        return name_node.value

    def get_modlist_declaration(self) -> ast.List:
        for keyword in self.pkg_node.keywords:
            if keyword.arg == "mods":
                modlist = keyword.value
                if not isinstance(modlist, ast.List):
                    raise PackageFileError(
                        "`mods` keyword argument must be a list statement."
                    )
                for mod in modlist.elts:
                    if not self._refs.is_mod_target(mod):
                        raise PackageFileError(
                            f"'{ast.unparse(mod)}' is not a valid `glia.Mod` declaration."
                        )
                return modlist
        raise PackageFileError("Missing `mods` keyword argument for `Package` call.")

    def get_mods(self) -> list[Mod]:
        mods = []
        for mod_call in self.get_modlist_declaration().elts:
            mod_call: ast.Call
            if any(
                not isinstance(arg, ast.Constant)
                for arg in itertools.chain(
                    mod_call.args, (v.value for v in mod_call.keywords)
                )
            ):
                raise PackageFileError(
                    "Mod declarations may only use constant arguments."
                )
            mod_node = copy(mod_call)
            mod_node.func = ast.Name("Mod")
            mod_str = ast.unparse(mod_node)
            pkg_name = self.get_package_name()
            pkg = self.get_package_shim()
            try:
                mod = eval(mod_str, {"Mod": Mod})
                mod.set_package(pkg)
                mods.append(mod)
            except Exception as e:
                raise PackageFileError(
                    f"`{mod_str}` is not a valid glia.Mod declaration."
                ) from e
        return mods

    def add_mod(self, mod: Mod):
        modlist = self.get_modlist_declaration()
        modlist.elts.append(self.make_mod_node(mod))

    def make_mod_node(self, mod: Mod):
        call = self._refs.make_mod_call_node()
        for param in inspect.signature(Mod).parameters.values():
            try:
                value = getattr(mod, param.name)
            except AttributeError:
                raise PackageApiError(f"Can't read attribute `Mod.{param.name}`.")
            # Convert non-constants to string, otherwise `ast.Constant` calls `__repr__`
            # and can create invalid runtime code
            if type(value) not in (str, bool, int, float, type(None)):
                value = str(value)
            # Omit default parameter values
            if value == param.default:
                continue
            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                call.args.append(ast.Constant(value))
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                call.keywords.append(ast.keyword(param.name, ast.Constant(value)))
            else:
                raise PackageApiError("Couldn't process glia.Mod class arguments.")
        return call

    def write_in_place(self):
        self._path.write_text(
            black.format_file_contents(
                ast.unparse(self._module),
                fast=True,
                mode=black.Mode(),
            )
        )


def get_package_transformer(path: Path, pkg_id: str):
    module = ast.parse(path.read_text(), str(path))
    return PackageTransformer(path, module, pkg_id)


class NmodlWriter:
    def __init__(self, mod: Mod):
        self._mod = mod
        self._source = None
        self._driver: "NmodlDriver" = None

    @property
    def nmodl(self):
        try:
            import nmodl
            import nmodl.ast
            import nmodl.dsl
            import nmodl.visitor

            return nmodl
        except ImportError as e:
            raise RuntimeError(
                "`nmodl` package unavailable. Try to install it. If you are on Windows, "
                "please see https://github.com/BlueBrain/nmodl/issues/1112"
            ) from e

    def parse_source(self, source: Path, dialect: SupportedDialect = None):
        text = source.read_text()
        for patch in get_patches(dialect):
            text = patch.common(text)
        self._source = self.nmodl.NmodlDriver().parse_string(text)

    def update_suffix_ast(self, name=None):
        if name is None:
            name = self._mod.mod_name
        self._check_source()
        visitor = self.nmodl.visitor.AstLookupVisitor()
        block = visitor.lookup(self._source, self.nmodl.ast.AstNodeType.NEURON_BLOCK)
        statements = [
            st for st in block[0].statement_block.statements if not st.is_suffix()
        ]
        if self._mod.is_artificial_cell:
            suffix = "ARTIFICIAL_CELL"
        elif self._mod.is_point_process:
            suffix = "POINT_PROCESS"
        else:
            suffix = "SUFFIX"

        statements.insert(
            0,
            self.nmodl.ast.Suffix(
                self.nmodl.ast.Name(self.nmodl.ast.String(suffix)),
                self.nmodl.ast.Name(self.nmodl.ast.String(name)),
            ),
        )
        block[0].statement_block.statements = statements

    def import_source(self, source: Path, dest: Path, dialect: SupportedDialect = None):
        mod_path = (dest / self._mod.relpath).resolve()
        mod_path.parent.mkdir(parents=True, exist_ok=True)
        # Import dialect source
        self.parse_source(source, dialect=dialect)
        self.update_suffix_ast()
        # Write common source
        self.write(mod_path)

    def write(self, target: Path, dialect: SupportedDialect = None):
        self._check_source()
        text = self.nmodl.to_nmodl(self._source)
        for patch in get_patches(dialect):
            text = patch.dialect(text)
        target.write_text(text)

    def extract_source_info(self):
        self._check_source()
        visitor = self.nmodl.visitor.AstLookupVisitor()
        suffixes = visitor.lookup(self._source, self.nmodl.ast.AstNodeType.SUFFIX)
        if len(suffixes) > 1:
            raise ModSourceError(
                "Multiple SUFFIX, POINT_PROCESS, or ARTIFICIAL_CELL statements detected."
            )
        elif not suffixes:
            self._mod.is_artificial_cell = False
            self._mod.is_point_process = False
        else:
            value = str(suffixes[0].type.value)
            self._mod.is_artificial_cell = value == "ARTIFICIAL_CELL"
            self._mod.is_point_process = value == "POINT_PROCESS"

    def _check_source(self):
        if self._source is None:
            raise RuntimeError(f"No nmodl source set for {self._mod}")


_patches = defaultdict(list)


class DialectPatch(abc.ABC):
    @abc.abstractmethod
    def dialect(self, common_text: str) -> str:
        pass

    @abc.abstractmethod
    def common(self, dialect_text: str) -> str:
        pass

    def __init_subclass__(cls, dialect: SupportedDialect, **kwargs):
        super().__init_subclass__(**kwargs)
        register_patch(dialect, cls())


def register_patch(dialect: SupportedDialect, patch: DialectPatch):
    _patches[dialect].append(patch)


def get_patches(dialect: SupportedDialect) -> list[DialectPatch]:
    return _patches[dialect]


class UniDirEqPatch(DialectPatch, dialect="arbor"):
    """
    Replace arbor's `a <-> (0, f)` with NMODL's `a << f`.
    """

    dialect_regex = re.compile(r"~\s*(\w+)\s*<->\s*\(\s*0\s*,\s*(.+)\)")
    common_regex = re.compile(r"~\s*(\w+)\s*<<\s*\((.+)\)")

    def dialect(self, common_text: str) -> str:
        return self.common_regex.sub(r"~ \1 <-> (0, \2)", common_text)

    def common(self, dialect_text: str) -> str:
        return self.dialect_regex.sub(r"~ \1 << (\2)", dialect_text)


class IndependentPatch(DialectPatch, dialect="neuron"):
    """
    Strip INDEPENDENT blocks: does nothing and NMODL parses them incorrectly.
    """

    regex = re.compile(r"INDEPENDENT\s*\{.+?\}", re.DOTALL)

    def dialect(self, common_text: str) -> str:
        return common_text

    def common(self, dialect_text: str) -> str:
        return self.regex.sub("", dialect_text)



from typing import Type

from errr.tree import exception as _e
from errr.tree import make_tree as _make_tree

_make_tree(
    globals(),
    GliaError=_e(
        CompileError=_e(),
        LibraryError=_e(),
        NeuronError=_e(),
        ResolveError=_e(
            TooManyMatchesError=_e("matches", "asset", "pkg", "variant"),
            NoMatchesError=_e("pkg", "variant"),
            UnknownAssetError=_e(),
            AssetLookupError=_e(),
        ),
        PackageError=_e(
            PackageApiError=_e(),
            PackageFileError=_e(),
            PackageModError=_e(),
            PackageProjectError=_e(),
            PackageVersionError=_e(),
        ),
        CatalogueError=_e(BuildCatalogueError=_e()),
        NmodlError=_e(ModSourceError=_e()),
    ),
)

GliaError: Type[Exception]
CompileError: Type[GliaError]
LibraryError: Type[GliaError]
NeuronError: Type[GliaError]
ResolveError: Type[GliaError]
TooManyMatchesError: Type[ResolveError]
NoMatchesError: Type[ResolveError]
UnknownAssetError: Type[ResolveError]
AssetLookupError: Type[ResolveError]
PackageError: Type[GliaError]
PackageApiError: Type[PackageError]
PackageFileError: Type[PackageError]
PackageModError: Type[PackageError]
PackageProjectError: Type[PackageError]
PackageVersionError: Type[PackageError]
CatalogueError: Type[GliaError]
BuildCatalogueError: Type[CatalogueError]
NmodlError: Type[GliaError]
ModSourceError: Type[NmodlError]


class PackageWarning(Warning):
    pass



import os
import unittest

from patch import p

import glia._glia
from glia._fs import read_cache
from tests._shared import skipUnlessTestMods


@skipUnlessTestMods
class TestPackageDiscovery(unittest.TestCase):
    """
    Check if packages can be discovered.
    """

    def test_discovery(self):
        self.assertGreater(len(glia._manager.packages), 0)


@skipUnlessTestMods
class TestCompilation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        glia.compile()

    def test_caching(self):
        # Test whether the compilation was cached
        cache = read_cache()
        pkg = glia._manager.package("glia_test_mods")
        self.assertEqual(pkg.mod_hash, cache["mod_hashes"].get(pkg.hash))

    def test_compilation(self):
        # Check that the library files exist
        for path in glia._manager.get_libraries():
            with self.subTest(path=path):
                self.assertTrue(os.path.exists(path), "Missing library file")

    def test_insert(self):
        # Test mechanism insertion
        mech = glia._manager.insert(p.Section(), "Na")
        self.assertIsInstance(mech, glia._glia.MechAccessor)
        self.assertTrue(glia._manager.test_mechanism("AMPA"))
        self.assertTrue(glia._manager.test_mechanism("Na"))

    def test_insert_attrs(self):
        # Test mechanism attributes
        sec = p.Section()
        mech = glia._manager.insert(sec, "Na", attributes={"gnabar": 30})
        self.assertEqual(30, mech.get_parameter("gnabar"), "Attribute not set")
        self.assertRaises(
            AttributeError,
            glia._manager.insert,
            p.Section(),
            "Na",
            attributes={"doesntexist": 30},
        )

    def test_insert_pp(self):
        # Test point process insertion
        pp = glia._manager.insert(p.Section(), "AMPA")
        self.assertIsInstance(pp, glia._glia.MechAccessor)

    def test_insert_pp_attrs(self):
        glia._manager.insert(p.Section(), "AMPA", attributes={"U": 30})
        self.assertRaises(
            AttributeError,
            glia._manager.insert,
            p.Section(),
            "AMPA",
            attributes={"doesntexist": 30},
        )


class TestBuiltins(unittest.TestCase):
    def test_builtins(self):
        from patch import p

        import glia

        s = p.Section()
        glia.insert(s, "pas")
        for pkg in glia._manager.packages:
            if pkg.name == "NEURON":
                nrn_pkg = pkg
                break
        else:
            self.fail("NEURON builtin package not found.")
        self.assertEqual(
            [m.asset_name for m in nrn_pkg.mods],
            [
                "APCount",
                "AlphaSynapse",
                "Exp2Syn",
                "ExpSyn",
                "IClamp",
                "OClamp",
                "PointProcessMark",
                "SEClamp",
                "VClamp",
                "extracellular",
                "fastpas",
                "hh",
                "k_ion",
                "na_ion",
                "pas",
            ],
            "NEURON builtins incorrect",
        )
